// [[file:../arroyo-native-parser.org::*The Arroyo HTML exporter][The Arroyo HTML exporter:1]]
use anyhow::Result;
use pyo3::prelude::*;
use std::collections::{HashMap, HashSet};
use std::convert::From;
use std::fs;
use std::io::{Error, Write};
use std::iter;
use std::marker::PhantomData;

use regex::Regex;

use orgize::export::{DefaultHtmlHandler, HtmlEscape, HtmlHandler, SyntectHtmlHandler};
use orgize::{Element, Org};
// The Arroyo HTML exporter:1 ends here

// [[file:../arroyo-native-parser.org::*The Arroyo HTML exporter][The Arroyo HTML exporter:2]]
#[derive(Default, Debug, Clone)]
#[pyclass(dict)]
pub struct ExportOptions {
    /// id:{the_id} -> URL rewrites
    #[pyo3(get)]
    pub link_retargets: HashMap<String, String>,
    #[pyo3(get)]
    pub ignore_tags: HashSet<String>,
    #[pyo3(get)]
    pub limit_headings: HashSet<String>,
    #[pyo3(get)]
    pub include_subheadings: bool,
}

#[pymethods]
impl ExportOptions {
    #[new]
    fn new(
        link_retargets: HashMap<String, String>,
        ignore_tags: Vec<String>,
        limit_headings: Vec<String>,
        include_subheadings: Option<bool>,
    ) -> Self {
        let mut lh2 = HashSet::new();
        lh2.extend(limit_headings);
        let mut tags = HashSet::new();
        tags.extend(ignore_tags);
        ExportOptions {
            link_retargets,
            limit_headings: lh2,
            ignore_tags: tags,
            include_subheadings: include_subheadings.unwrap_or(false),
            ..Default::default()
        }
    }

    // add these later if i feel like it
    // pub fn __repr__(slf: PyRef<'_, Self>) -> PyResult<String> {
    //     Ok(slf.to_string())
    // }

    // pub fn __str__(slf: PyRef<'_, Self>) -> PyResult<String> {
    //     Self::__repr__(slf)
    // }
}
// The Arroyo HTML exporter:2 ends here

// [[file:../arroyo-native-parser.org::*The Arroyo HTML exporter][The Arroyo HTML exporter:3]]
pub struct ArroyoHtmlHandler<E: From<Error>, H: HtmlHandler<E>> {
    pub options: ExportOptions,
    /// inner html handler
    pub inner: H,
    /// handler error type
    pub error_type: PhantomData<E>,
    /// file-property drawer state tracking
    current_drawer: Option<String>,
    in_public_heading: bool,
    ignore_deeper_than: usize,
    heading_breadcrumbs: Vec<String>,
}

impl<E: From<Error>, H: HtmlHandler<E>> ArroyoHtmlHandler<E, H> {
    pub fn new(options: ExportOptions, inner: H) -> Self {
        let iph = options.clone().limit_headings.is_empty();
        ArroyoHtmlHandler {
            inner,
            options,
            in_public_heading: iph,
            ..Default::default()
        }
    }

    pub fn rewrite_link_from(&self, id: &String) -> String {
        match self.options.link_retargets.get(id) {
            Some(path) => HtmlEscape(&path).to_string(),
            _ => HtmlEscape(format!("/404?key={}", id)).to_string(),
        }
    }
}

impl<E: From<Error>, H: HtmlHandler<E>> Default for ArroyoHtmlHandler<E, H> {
    fn default() -> Self {
        ArroyoHtmlHandler {
            inner: H::default(),
            error_type: PhantomData,
            current_drawer: None,
            in_public_heading: false,
            heading_breadcrumbs: vec![],
            ignore_deeper_than: 999, // please don't make a 1000 deep heading as the first one in the doc...

            options: ExportOptions::default(),
        }
    }
}
// The Arroyo HTML exporter:3 ends here

// [[file:../arroyo-native-parser.org::*The Custom HTML Exporter Extensions][The Custom HTML Exporter Extensions:1]]
impl<E: From<Error>, H: HtmlHandler<E>> HtmlHandler<E> for ArroyoHtmlHandler<E, H> {
    fn start<W: Write>(&mut self, mut w: W, element: &Element) -> Result<(), E> {
        match &self.current_drawer {
            None => {}
            Some(drawer_name) => {
                if vec![
                    String::from("PROPERTIES"),
                    String::from("REVIEW_DATA"),
                    String::from("LOGBOOK"),
                ]
                .contains(&drawer_name)
                {
                    return Ok(());
                }
            }
        };
        // if !self.in_public_heading {
        //     return Ok(());
        // }
        match element {
// The Custom HTML Exporter Extensions:1 ends here

// [[file:../arroyo-native-parser.org::*The Custom HTML Exporter Extensions][The Custom HTML Exporter Extensions:2]]
            Element::Title(title) => {
                let properties = title.properties.clone().into_hash_map();
                let our_new_id = properties.get("ID");
                let our_level = title.level;

                if self.heading_breadcrumbs.len() < our_level {
                    let diff = our_level - self.heading_breadcrumbs.len();
                    self.heading_breadcrumbs
                        .extend(iter::repeat("SENTINEL".into()).take(diff));
                }
                self.heading_breadcrumbs.truncate(our_level);
                match our_new_id.clone() {
                    None => {}
                    Some(id) => {
                        self.heading_breadcrumbs.push(id.to_string());
                    }
                }

                let has_ignore_tag = title
                    .tags
                    .clone()
                    .into_iter()
                    .map(String::from)
                    .find(|v| self.options.ignore_tags.contains(v))
                    .is_some();
                // if it's a new ignore tag or a subheading of one we already parsed
                self.in_public_heading = !has_ignore_tag && title.level <= self.ignore_deeper_than;
                self.in_public_heading = self.in_public_heading
                    && self.options.limit_headings.is_empty()
                    || self
                        .options
                        .limit_headings
                        .intersection(&HashSet::from_iter(
                            self.heading_breadcrumbs.clone().into_iter(),
                        ))
                        .count()
                        != 0;
                if self.in_public_heading {
                    self.ignore_deeper_than = 999;
                }
                self.in_public_heading;

                if has_ignore_tag {
                    self.ignore_deeper_than = title.level;
                }

                let mut parts = Vec::new();
                parts.push(format!(
                    "<h{}>",
                    if title.level <= 6 { title.level } else { 6 }
                ));

                let props = title.properties.clone().into_hash_map();
                let id = props.get("ID");
                match id {
                    Some(id) => parts.push(format!("<a class=\"id-anchor\" id=\"{}\"></a>", id)),
                    None => {}
                }

                match &title.keyword {
                    Some(kw) => parts.push(format!("<span class=\"task task-{}\"></span>", kw)),
                    None => {}
                }

                (title.tags)
                    .iter()
                    .for_each(|tag| parts.push(format!("<span class=\"tag tag-{}\"></span>", tag)));

                if self.in_public_heading {
                    write!(w, "{}", parts.join(""))?
                }
            }
// The Custom HTML Exporter Extensions:2 ends here

// [[file:../arroyo-native-parser.org::*The Custom HTML Exporter Extensions][The Custom HTML Exporter Extensions:3]]
            Element::Drawer(drawer) => {
                self.current_drawer = Some(drawer.name.to_string());
            }
// The Custom HTML Exporter Extensions:3 ends here

// [[file:../arroyo-native-parser.org::*The Custom HTML Exporter Extensions][The Custom HTML Exporter Extensions:4]]
            Element::Text { value: before } => {
                let re = Regex::new(
                    r"(?x)
                          \{\{([^\}]+)}     # grab the answer
                          (?:\{([^\}]+?)})? # grab a hint
                          @[0-9]+}          # exclude the question number
                        ",
                )
                .unwrap();
                if self.in_public_heading {
                    let after =
                        re.replace_all(before, "<span class='fc-cloze' title='$2'>$1</span>");
                    if after.eq(before) {
                        self.inner.start(w, &Element::Text { value: after })?
                    } else {
                        write!(w, "{}", after)?
                    }
                }
            }
// The Custom HTML Exporter Extensions:4 ends here

// [[file:../arroyo-native-parser.org::*The Custom HTML Exporter Extensions][The Custom HTML Exporter Extensions:5]]
            Element::Link(link) => {
                let string_path = link.path.to_string();
                let (proto, stripped_dest) = match string_path.split_once(':') {
                    Some((proto, stripped_dest)) => (proto, stripped_dest.into()),
                    None => ("", string_path.clone()),
                };
                let mut desc =
                    HtmlEscape(link.desc.clone().unwrap_or(link.path.clone())).to_string();

                if self.in_public_heading {
                    let image_maybe = vec![".png", ".svg", ".jpg", ".jpeg"];
                    if image_maybe
                        .iter()
                        .any(|&suffix| string_path.ends_with(suffix))
                    {
                        desc = format!("<img src=\"{}\" alt=\"{}\"/>", string_path, desc);
                        write!(w, "<a href=\"{}\">{}</a>", string_path, desc,)?
                    }

                    match proto {
                        "id" => write!(
                            w,
                            "<a href=\"{}\">{}</a>",
                            self.rewrite_link_from(&stripped_dest),
                            desc,
                        )?,
                        "roam" => write!(
                            w,
                            "<a href=\"/404?key={}\">{}</a>",
                            HtmlEscape(&link.path),
                            desc,
                        )?,
                        _ => self.inner.start(w, &Element::Link(link.clone()))?,
                    }
                }
            }
// The Custom HTML Exporter Extensions:5 ends here

// [[file:../arroyo-native-parser.org::*The Custom HTML Exporter Extensions][The Custom HTML Exporter Extensions:6]]
            Element::SourceBlock(block) => {
                if self.in_public_heading {
                    let args = String::from(block.arguments.clone());
                    if args.len() > 0 {
                        write!(w, "<span class='babel-args'>{}</span>", args)?
                    }
                    self.inner.start(w, &Element::SourceBlock(block.clone()))?
                }
            }
// The Custom HTML Exporter Extensions:6 ends here

// [[file:../arroyo-native-parser.org::*The Custom HTML Exporter Extensions][The Custom HTML Exporter Extensions:7]]
            _ => {
                if self.in_public_heading {
                    self.inner.start(w, element)?
                }
            }
        }
        Ok(())
    }

    fn end<W: Write>(&mut self, w: W, element: &Element) -> Result<(), E> {
        match element {
            // reset the drawer state tracking
            Element::Drawer(_drawer) => {
                self.current_drawer = None;
            }
            _ => {
                if !self.in_public_heading {
                    return Ok(());
                }
                if self.current_drawer.is_some() {
                    return Ok(());
                }
                self.inner.end(w, element)?
            }
        }
        Ok(())
    }
}
// The Custom HTML Exporter Extensions:7 ends here

// [[file:../arroyo-native-parser.org::*The API Interface][The API Interface:1]]
// sure would be nice..... some day i'll understand lifetimes enough
// to write a function that goes path -> orgize::Org
// use crate::parse::orgize_document;

pub fn htmlize_file(path: String, options: ExportOptions) -> Result<String> {
    let mut handler = ArroyoHtmlHandler::new(options, SyntectHtmlHandler::new(DefaultHtmlHandler));
    let org = String::from_utf8(fs::read(path.clone())?).unwrap();
    let org_tree = &Org::parse_custom(
        &org,
        &orgize::ParseConfig {
            // Need to pull these from environment or options...
            todo_keywords: (
                vec![
                    "NEXT".to_string(),
                    "INPROGRESS".to_string(),
                    "WAITING".to_string(),
                ],
                vec!["DONE".to_string(), "CANCELLED".to_string()],
            ),
            ..Default::default()
        },
    );

    let mut vec = vec![];
    org_tree.write_html_custom(&mut vec, &mut handler)?;
    Ok(String::from_utf8(vec)?)
}
// The API Interface:1 ends here
